"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Dependency = exports.DependencyType = void 0;
const blueimp_md5_1 = __importDefault(require("blueimp-md5"));
const events_1 = __importDefault(require("events"));
const greybel_core_1 = require("greybel-core");
const greybel_transpiler_1 = require("greybel-transpiler");
const greyscript_core_1 = require("greyscript-core");
var DependencyType;
(function (DependencyType) {
    DependencyType[DependencyType["Main"] = 0] = "Main";
    DependencyType[DependencyType["Import"] = 1] = "Import";
    DependencyType[DependencyType["Include"] = 2] = "Include";
    DependencyType[DependencyType["NativeImport"] = 3] = "NativeImport";
})(DependencyType = exports.DependencyType || (exports.DependencyType = {}));
class Dependency extends events_1.default {
    constructor(options) {
        super();
        const me = this;
        me.target = options.target;
        me.id = (0, blueimp_md5_1.default)(options.target);
        me.resourceHandler = options.resourceHandler;
        me.chunk = options.chunk;
        me.dependencies = new Set();
        me.type = options.type || DependencyType.Main;
        me.context = options.context;
        me.ref = options.ref;
        const namespace = me.context.createModuleNamespace(me.id);
        const resourceDependencyMap = me.context.getOrCreateData(greybel_transpiler_1.ContextDataProperty.ResourceDependencyMap, () => new Map());
        resourceDependencyMap.set(`${namespace}:${me.type}`, me);
        me.context.getOrCreateData(greybel_transpiler_1.ContextDataProperty.DependencyCallStack, () => []);
    }
    getId() {
        return this.id;
    }
    getNamespace() {
        const me = this;
        return me.context.modules.get(me.id);
    }
    fetchNativeImports() {
        const me = this;
        const result = [];
        for (const item of me.dependencies) {
            if (item.type === DependencyType.NativeImport) {
                (0, greybel_transpiler_1.merge)(result, [...item.fetchNativeImports()]);
                result.push(item);
            }
        }
        return new Set(result);
    }
    resolve(path, type, ref) {
        return __awaiter(this, void 0, void 0, function* () {
            const me = this;
            const context = me.context;
            const { modules } = context;
            const resourceDependencyMap = context.get(greybel_transpiler_1.ContextDataProperty.ResourceDependencyMap);
            const resourceHandler = me.resourceHandler;
            const subTarget = yield resourceHandler.getTargetRelativeTo(me.target, path);
            const id = (0, blueimp_md5_1.default)(subTarget);
            const namespace = modules.get(id);
            if (resourceDependencyMap.has(`${namespace}:${type}`)) {
                return resourceDependencyMap.get(`${namespace}:${type}`);
            }
            if (!(yield resourceHandler.has(subTarget))) {
                throw new Error('Dependency ' + subTarget + ' does not exist...');
            }
            const content = yield resourceHandler.get(subTarget);
            me.emit('parse-before', subTarget);
            try {
                const parser = new greyscript_core_1.Parser(content, {
                    filename: subTarget
                });
                const chunk = parser.parseChunk();
                const dependency = new Dependency({
                    target: subTarget,
                    resourceHandler,
                    chunk,
                    type,
                    context,
                    ref
                });
                me.emit('parse-after', dependency);
                return dependency;
            }
            catch (err) {
                if (err instanceof greybel_transpiler_1.BuildError) {
                    throw err;
                }
                throw new greybel_transpiler_1.BuildError(err.message, {
                    target: subTarget,
                    range: err.range
                });
            }
        });
    }
    findInjections() {
        return __awaiter(this, void 0, void 0, function* () {
            const me = this;
            const { injects } = me.chunk;
            const injections = new Map();
            for (const item of injects) {
                const injectionTarget = yield me.resourceHandler.getTargetRelativeTo(me.target, item.path);
                if (!(yield me.resourceHandler.has(injectionTarget))) {
                    throw new Error('Injection ' + injectionTarget + ' does not exist...');
                }
                const content = yield me.resourceHandler.get(injectionTarget);
                injections.set(item.path, content);
            }
            me.injections = injections;
            return injections;
        });
    }
    findDependencies() {
        return __awaiter(this, void 0, void 0, function* () {
            const me = this;
            const { imports, includes, nativeImports } = me.chunk;
            const sourceNamespace = me.getNamespace();
            const dependencyCallStack = me.context.get(greybel_transpiler_1.ContextDataProperty.DependencyCallStack);
            const namespaces = [...(0, greybel_transpiler_1.fetchNamespaces)(me.chunk)];
            const literals = [...me.chunk.literals];
            const result = [];
            dependencyCallStack.push(sourceNamespace);
            // handle native imports
            for (const nativeImport of nativeImports) {
                if (!nativeImport.eval || !nativeImport.emit) {
                    continue;
                }
                const dependency = yield me.resolve(nativeImport.directory, DependencyType.NativeImport, nativeImport);
                const namespace = dependency.getNamespace();
                if (dependencyCallStack.includes(namespace)) {
                    throw new Error(`Circular dependency from ${me.target} to ${dependency.target} detected.`);
                }
                const relatedDependencies = yield dependency.findDependencies();
                (0, greybel_transpiler_1.merge)(namespaces, relatedDependencies.namespaces);
                (0, greybel_transpiler_1.merge)(literals, relatedDependencies.literals);
                result.push(dependency);
            }
            // handle internal includes/imports
            const items = [...imports, ...includes];
            for (const item of items) {
                const type = item instanceof greybel_core_1.ASTFeatureIncludeExpression
                    ? DependencyType.Include
                    : DependencyType.Import;
                const dependency = yield me.resolve(item.path, type, item);
                const namespace = dependency.getNamespace();
                if (dependencyCallStack.includes(namespace)) {
                    throw new Error(`Circular dependency from ${me.target} to ${dependency.target} detected.`);
                }
                const chunk = dependency.chunk;
                item.chunk = chunk;
                item.namespace = namespace;
                const relatedDependencies = yield dependency.findDependencies();
                (0, greybel_transpiler_1.merge)(result, [...dependency.fetchNativeImports()]);
                (0, greybel_transpiler_1.merge)(namespaces, relatedDependencies.namespaces);
                (0, greybel_transpiler_1.merge)(literals, relatedDependencies.literals);
                result.push(dependency);
            }
            yield this.findInjections();
            const dependencies = new Set(result);
            me.dependencies = dependencies;
            dependencyCallStack.pop();
            return {
                dependencies,
                namespaces,
                literals
            };
        });
    }
}
exports.Dependency = Dependency;
